import os
import re
import argparse
import uproot
import numpy as np

from lhcb_ftcalib._version import __version__
from lhcb_ftcalib.TaggerCollection import TaggerCollection
from lhcb_ftcalib.calibration_functions import PolynomialCalibration
from lhcb_ftcalib.save_calibration import save_calibration
from lhcb_ftcalib.apply_tagger import TargetTaggerCollection, get_link_by_name
from lhcb_ftcalib.printing import (warning, info, FTCalibException, section_header,
                                   print_tagger_statistics, print_tagger_performances,
                                   MissingBranch, MissingFile, MissingTree)


def main():
    try:
        args = parse_args()
        section_header(f"lhcb_ftcalib version {__version__}")
        loadplan = read_file(args.rootfile,
                             ID        = args.id_branch,
                             TAGGERS   = args.taggers,
                             SSTAGGERS = args.SStaggers,
                             OSTAGGERS = args.OStaggers,
                             TAU       = args.tau,
                             TAUERR    = args.tauerr,
                             WEIGHT    = args.weights,
                             SEL       = args.selection,
                             ask       = args.interactive)
    except FTCalibException as ex:
        print(ex)
        return
    if loadplan is None:
        print("Aborted")
        return

    run(args, load_data(args, loadplan), loadplan)


def parse_args():
    parser = argparse.ArgumentParser(description="LHCb Flavour Tagging calibration software")
    parser.add_argument("rootfile",  type=str, help="ROOT file and tree to read tagging data from. (example: \"file.root:DecayTree\")")
    parser.add_argument("-t",         dest="taggers",    type=str, nargs="+", default=[], help="Enumeration of taggers to find in file. This argument will try to match partial names, e.g. \"MuonLatest\"->\"B_OSMuonLatest_TAGDEC/ETA\"")
    parser.add_argument("-SS",        dest="SStaggers",  type=str, nargs="+", default=[], help="Enumeration of same side taggers to find in file. This argument will try to match partial names, e.g. \"MuonLatest\"->\"B_OSMuonLatest_TAGDEC/ETA\"")
    parser.add_argument("-OS",        dest="OStaggers",  type=str, nargs="+", default=[], help="Enumeration of opposite side taggers to find in file. This argument will try to match partial names, e.g. \"MuonLatest\"->\"B_OSMuonLatest_TAGDEC/ETA\"")
    parser.add_argument("-id",        dest="id_branch",  type=str, help="Name of the B meson id branch")
    parser.add_argument("-mode",      dest="mode",       type=str, choices=["Bu", "Bd", "Bs"],  help="Calibration mode")
    parser.add_argument("-tau",       dest="tau",        type=str, help="Name of the decay time branch")
    parser.add_argument("-tauerr",    dest="tauerr",     type=str, help="Name of the decay time uncertainty branch")
    parser.add_argument("-timeunit",  dest="timeunit",   type=str, default="ns", choices=["ns", "ps", "fs"], help="Decay time unit")
    parser.add_argument("-weights",   dest="weights",    type=str, help="Name of the per-event weight branch")
    parser.add_argument("-op",        dest="op",         type=str, required=True,  nargs='+', choices=["calibrate", "combine", "apply"], help="What to do with the loaded taggers")
    parser.add_argument("-write",     dest="write",      type=str, help="Name of a root file where to store calibrated branches")
    parser.add_argument("-selection", dest="selection",  type=str, help=r"Selection expression (example: 'eventNumber%%2==0')")
    parser.add_argument("-input_cal", dest="input_cal",  type=str, help="JSON file of the input calibrations")
    parser.add_argument("-plot",      dest="plot",       action="store_true", help="If set, plots calibration curves")

    parser.add_argument("-fun",  type=str, dest="fun", nargs=2, default=["POLY", 1], help="CalibrationFunction followed by degree (default: %(default)s)")
    parser.add_argument("-link", type=str, dest="link", choices=["mistag", "logit", "rlogit", "probit", "rprobit", "cauchit", "rcauchit"], default="mistag", help="Link function (default: %(default)s)")

    parser.add_argument("-i", action="store_true", dest="interactive", help="Interactive mode, will ask for confirmation")
    # parser.add_argument("-filetype", type=str, default="root", choices=["root", "csv"], dest="filetype", help="Filetype to use instead of root files")

    args = parser.parse_args()

    if "apply" in args.op and args.input_cal is None:
        raise FTCalibException("You need to provide a file with input calibrations using the \"-input_cal\" argument")
    if args.input_cal is not None and not os.path.exists(args.input_cal):
        raise FTCalibException(f"Input calibration file {args.input_cal} does not exist")
    if args.id_branch is None and "calibrate" in args.op:
        raise FTCalibException("An ID branch is required for calibrations")

    return args


def assign_eta_omega(search_result):
    """ search_result is a list of two branches. This function
        identifies the mistag and tag dec branch """
    dec = next(b for b in search_result if b.endswith("DEC"))
    eta = next(b for b in search_result if b.endswith("ETA"))
    name = dec.replace("_TAGDEC", "").replace("_DEC", "").replace("_DECISION", "")
    return name, dec, eta


def get_selection_branches(selection, all_branches):
    info("Parsing selection", selection)

    def is_bool(match):
        return match in ("and", "or", "xor", "not")

    selection = selection.strip()
    functions = re.compile(r"[A-Za-z]+[_A-Za-z0-9]*\(")  # any identifier followed by '(' is a "function", not a branch
    selvars  = re.compile("[A-Za-z]+[_A-Za-z0-9]*")
    funcs = [match[:-1] for match in functions.findall(selection)]
    selvars = [match for match in selvars.findall(selection) if match not in funcs and not is_bool(match)]
    selvars = list(set(selvars))

    info("Selection branches: ", selvars)

    return selvars


def read_file(File, ID, TAGGERS=[], SSTAGGERS=[], OSTAGGERS=[], TAU=None, TAUERR=None, WEIGHT=None, SEL=None, ask=False):
    # Open file while anticipating all possible user errors
    loadplan = {"taggers"   : {},
                "sstaggers" : {},
                "ostaggers" : {},
                "branches"  : [],
                "selection" : SEL}
    if ":" in File:
        filename, treename = File.split(":")
    else:
        filename, treename = File, None

    if not os.path.exists(filename):
        raise MissingFile(f"File {filename} not found")

    tfile = uproot.open(filename)
    if treename is None:
        warning("TTree not specified, reading from first TTree in file")
        treename = tfile.keys()[0]
    else:
        if treename not in tfile:
            raise MissingTree(f"TTree \"{treename}\" not found in file {filename}. Available trees: {tfile.keys()}")
    tree = tfile[treename]
    info(f"Reading data from {filename}:{treename}")

    def add_if(branch):
        if branch is not None:
            if branch not in tree.keys():
                raise MissingBranch(f"Branch \"{branch}\" not found in {filename}:{treename}")
            loadplan["branches"].append(branch)

    add_if(ID)
    add_if(TAU)
    add_if(TAUERR)
    add_if(WEIGHT)

    # Get list of all possible taggers
    branches = tree.keys()
    if SEL is not None:
        loadplan["selection"] = SEL
        loadplan["sel_branches"] = get_selection_branches(SEL, branches)
        loadplan["branches"] += loadplan["sel_branches"]

    alltaggers = list(set(TAGGERS + SSTAGGERS + OSTAGGERS))

    for tagger_hint in alltaggers:
        search = [branch for branch in branches if tagger_hint in branch]
        search = [branch for branch in search if branch.endswith("ETA") or branch.endswith("DEC")]
        if len(search) != 2:
            raise FTCalibException(f"Too many or None of the branches in the file are matching the tagger hint \"{tagger_hint}\":\n\t* " + '\n\t* '.join(search))

        tagger, dec, eta = assign_eta_omega(search)
        if tagger_hint in TAGGERS and tagger not in loadplan["taggers"]:
            loadplan["taggers"][tagger] = {"dec": dec, "mistag": eta}
        if tagger_hint in SSTAGGERS and tagger not in loadplan["sstaggers"]:
            loadplan["sstaggers"][tagger] = {"dec": dec, "mistag": eta}
        if tagger_hint in OSTAGGERS and tagger not in loadplan["ostaggers"]:
            loadplan["ostaggers"][tagger] = {"dec": dec, "mistag": eta}

        loadplan["branches"].append(dec)
        loadplan["branches"].append(eta)

    info("The following taggers have been found")
    for k, v in loadplan["taggers"].items():
        print(f"\t-> {k} =", v["mistag"], "&", v["dec"])
    for k, v in loadplan["sstaggers"].items():
        print(f"\t-> {k} =", v["mistag"], "&", v["dec"])
    for k, v in loadplan["ostaggers"].items():
        print(f"\t-> {k} =", v["mistag"], "&", v["dec"])

    if ask:
        if not (input("Continue? <y/n> ") in ("Y", "y", "Yes", "yes")):
            return None

    loadplan["ID"]       = ID
    loadplan["TAU"]      = TAU
    loadplan["TAUERR"]   = TAUERR
    loadplan["weight"]   = WEIGHT
    loadplan["tree"]     = treename
    loadplan["file"]     = filename
    loadplan["branches"] = list(set(loadplan["branches"]))  # There can be duplicate due to event selection

    return loadplan


def load_data(args, loadplan):
    data = uproot.open(loadplan["file"])[loadplan["tree"]].arrays(loadplan["branches"], library="pd")

    # Convert to picosenconds
    timefactor = {"ns" : 1e3, "ps" : 1.0, "fs" : "1e-3"}[args.timeunit]
    if loadplan["TAU"] is not None:
        data.loc[:, loadplan["TAU"]] *= timefactor
    if loadplan["TAUERR"] is not None:
        data.loc[:, loadplan["TAUERR"]] *= timefactor

    # pandas dataframe queries have problems with uint64 => convert to int64
    for c in data.columns.values:
        if data[c].dtype == np.uint64:
            # pandas eval has weird data type limitations
            warning(f"Branch {c} has type uint64. Converting to int64 for pandas")
            data[c] = data[c].astype(np.int64)

    return data


def write_branches(args, tc, filename, calibrated, append=False, write_extra_branches=False):
    import root_pandas  # uproot4 cannot write yet :(
    df = tc.get_dataframe(calibrated=calibrated)
    if write_extra_branches:
        df["ID"] = tc[0].stats.all_B_ID.copy()
        if tc[0].tau is not None:
            df["TAU"] = tc[0].tau
        if tc[0].tauerr is not None:
            df["TAUERR"] = tc[0].tauerr
    info(f"Writing file {filename}")
    if append:
        root_pandas.to_root(df, filename, key="TaggingTree", mode="a", store_index=False)
    else:
        if os.path.exists(filename):
            warning(f"File {filename} already exists => recreating")
        root_pandas.to_root(df, filename, key="TaggingTree", store_index=False)


def validate_ops(args):
    # Check whether order of requested calibration steps makes sense / is supported
    state = "___start"
    dfa = {
        "___start":      {"calibrate" : "calibrate", "apply" : "apply", "combine" : None},
        "apply":         {"calibrate" : None,        "apply" : None,    "combine" : "END"},
        "calibrate":     {"calibrate" : None,        "apply" : None,    "combine" : "combine"},
        "combine":       {"calibrate" : "END",       "apply" : None,    "combine" : None},
        "END":           {"calibrate" : None,        "apply" : None,    "combine" : None},
    }
    for i, op in enumerate(args.op):
        nextstate = dfa[state][op]
        if nextstate is None:
            if state == "___start":
                raise FTCalibException(f"Argument {i}: Procedure cannot begin with operation \"{op}\"")
            else:
                raise FTCalibException(f"Argument {i}: Operation \"{op}\" cannot come after \"{state}\"" + (" (nothing left to do)" if state == "END" else ""))
        state = nextstate


def run(args, data, loadplan):
    # import pudb
    # pu.db
    if loadplan["selection"] is not None:
        data.eval(f"FTCSELECTION={loadplan['selection']}", inplace=True)
        selection = data.FTCSELECTION.copy()
        rem = [b for b in loadplan["sel_branches"] if b not in loadplan["branches"]]
        data.drop(columns=["FTCSELECTION"] + rem, inplace=True)
    else:
        selection = None

    validate_ops(args)

    def get_tagger_collection(taggertype, apply_tagger):
        if len(loadplan[taggertype]) == 0:
            return None
        if apply_tagger:
            tc = TargetTaggerCollection()
            for name, tagbranch in loadplan[taggertype].items():
                tc.create_tagger(name,
                                 eta    = data[tagbranch["mistag"]],
                                 dec    = data[tagbranch["dec"]],
                                 weight = data[loadplan["weight"]] if loadplan["weight"] is not None else None)
        else:
            tc = TaggerCollection()
            for name, tagbranch in loadplan[taggertype].items():
                tc.create_tagger(name,
                                 eta_data  = data[tagbranch["mistag"]],
                                 dec_data  = data[tagbranch["dec"]],
                                 B_ID      = data[loadplan["ID"]],
                                 mode      = args.mode,
                                 selection = selection,
                                 tau_ps    = data[loadplan["TAU"]] if loadplan["TAU"] is not None else None,
                                 tauerr_ps = data[loadplan["TAUERR"]] if loadplan["TAUERR"] is not None else None,
                                 weight    = data[loadplan["weight"]] if loadplan["weight"] is not None else None)
        return tc

    tc_any, tc_ss, tc_os = (get_tagger_collection("taggers",   apply_tagger="apply" in args.op),
                            get_tagger_collection("sstaggers", apply_tagger="apply" in args.op),
                            get_tagger_collection("ostaggers", apply_tagger="apply" in args.op))

    calibration_filename = args.write + ".json" if args.write is not None else "out.json"
    calibration_rootfile = args.write + ".root" if args.write is not None else "out.root"
    stored_extra_branches = False

    def calibration_steps(tc):
        nonlocal stored_extra_branches

        if tc is not None:
            if args.fun != ["POLY", 1] or args.link != "mistag":
                tc.set_calibration(PolynomialCalibration(int(args.fun[1]) + 1, get_link_by_name(args.link)))
            tc.calibrate(ignore_eta_out_of_range=False)
            if args.plot:
                tc.plot_calibration_curves(savepath="/".join(args.write.split("/")[:-1]) if args.write is not None else ".")
            save_calibration(tc, title=calibration_filename)
            if args.write is not None:
                write_branches(args, tc, filename=calibration_rootfile, calibrated=True, append=True, write_extra_branches=not stored_extra_branches)
                stored_extra_branches = True  # Write ID, etc only once

    def combination_steps(tc, combination_name, apply_calib):
        if tc is not None:
            Combination = tc.combine_taggers(combination_name, calibrated=True)
            return Combination

    def apply_steps(tc):
        if tc is not None:
            tc.load_calibrations(args.input_cal)
            tc.apply()
            write_branches(args, tc, filename=calibration_rootfile, calibrated=True, append=True)

    # Workflows for applying taggers and calibrating them are forced to be separate
    if "apply" in args.op:
        combinations = None
        for i, operation in enumerate(args.op):
            if operation == "apply":
                apply_steps(tc_any)
                apply_steps(tc_ss)
                apply_steps(tc_os)
            if operation == "combine":
                combinations = TargetTaggerCollection()
                any_comb = combination_steps(tc_any, combination_name="Combination", apply_calib=True)
                if any_comb is not None:
                    combinations.add_taggers(any_comb)
                ss_comb = combination_steps(tc_ss,  combination_name="SS_Combination", apply_calib=True)
                if ss_comb is not None:
                    combinations.add_taggers(ss_comb)
                os_comb = combination_steps(tc_os,  combination_name="OS_Combination", apply_calib=True)
                if os_comb is not None:
                    combinations.add_taggers(os_comb)
                print_tagger_statistics(combinations, calibrated=False, selected=False)
                write_branches(args, combinations, filename=calibration_rootfile, calibrated=False, append=True)
                combinations.load_calibrations(args.input_cal)
                combinations.apply()  # Apply directly
                print_tagger_statistics(combinations, calibrated=True, selected=False)
                write_branches(args, combinations, filename=calibration_rootfile, calibrated=True, append=True)
    else:
        # Execute calibration and combination steps
        combinations = None
        for i, operation in enumerate(args.op):
            if operation == "calibrate":
                if combinations is not None:
                    # Calibrate all the combinations individually
                    calibration_steps(combinations)
                else:
                    calibration_steps(tc_any)
                    calibration_steps(tc_ss)
                    calibration_steps(tc_os)
            if operation == "combine":
                combinations = TaggerCollection()
                any_comb = combination_steps(tc_any, combination_name="Combination", apply_calib=False)
                if any_comb is not None:
                    combinations.add_taggers(any_comb)
                ss_comb  = combination_steps(tc_ss,  combination_name="SS_Combination", apply_calib=False)
                if ss_comb is not None:
                    combinations.add_taggers(ss_comb)
                os_comb  = combination_steps(tc_os,  combination_name="OS_Combination", apply_calib=False)
                if os_comb is not None:
                    combinations.add_taggers(os_comb)
                print_tagger_statistics(combinations, calibrated=False, selected=False)
                write_branches(args, combinations, filename=calibration_rootfile, calibrated=False, append=True)


if __name__ == "__main__":
    main()
